@@ -163,12 +163,21 @@ class Trainer(CommMixin):
         train_batch_per_replica = self.config.train.train_batch_per_replica
         dp_shard_size = self.config.policy.parallelism.dp_shard_size
         error_msg = f"train_batch_per_replica({train_batch_per_replica}) of {policy_type} must be divisible by dp_shard_size({dp_shard_size})"
+        mini_batch = self.config.train.train_policy.mini_batch
         if policy_type == "grpo":
-            mini_batch = self.config.train.train_policy.mini_batch
             error_msg += f" * mini_batch({mini_batch})"
-        assert dp_shard_size == self.parallel_dims.dp_shard
-        assert dp_shard_size > 0, "dp_shard_size must be greater than 0"
-        assert train_batch_per_replica % (dp_shard_size * mini_batch) == 0, error_msg
+            assert dp_shard_size == self.parallel_dims.dp_shard
+            assert dp_shard_size > 0, "dp_shard_size must be greater than 0"
+            assert (
+                train_batch_per_replica % (dp_shard_size * mini_batch) == 0
+            ), error_msg
+        else:
+            # TODO(jiaxinc): Optimize this:
+            #  for SFT,`train_batch_per_replica` stands for the batch_size for a DP worker,
+            #  not really for a training replica
+            assert (
+                train_batch_per_replica % mini_batch == 0
+            ), f"train_batch_per_replica({train_batch_per_replica}) of {policy_type} must be divisible by mini_batch({mini_batch})"
         logger.info("Config checked successfully")
 
     @property
