@@ -73,6 +73,8 @@ from cosmos_rl.utils.pynccl import (
     create_nccl_uid,
     create_nccl_comm,
     nccl_send,
+    nccl_group_start,
+    nccl_group_end,
 )
 from cosmos_rl.utils.util import compute_logprobs as logprobs_computing
 
@@ -734,40 +736,63 @@ class GRPOTrainer(Trainer):
         # Here we use another comm to send weight to rollout
         # NCCL announces that multi-comm could lead to deadlocks if not synchronized
         with torch.cuda.stream(self.train_stream):
-            pre_P2R_collected_tensors: Dict[str, torch.Tensor] = (
-                self.pre_P2R_collect_parameters()
-            )
-            for insts_group in self.policy_to_rollout_insts:
-                for insts_for_per_param in insts_group.param_instructions:
-                    dest_name = insts_for_per_param.param_name
-                    for inst in insts_for_per_param.instructions:
-                        p_rank = inst.policy_rank
-                        r_rank = inst.rollout_rank
-                        tensor_split_strategys = inst.slice_strategy
-                        if dest_name not in self.map_w_from_policy_to_rollout:
-                            raise RuntimeError(
-                                f"dest_name {dest_name} not in map_w_from_policy_to_rollout"
-                            )
-                        local_view = self.map_w_from_policy_to_rollout[dest_name]
-                        if dest_name in pre_P2R_collected_tensors:
-                            local_view = pre_P2R_collected_tensors[dest_name]
-                        elif isinstance(local_view, Callable):
-                            local_view = local_view()
-                        else:
-                            pass
+            with torch.no_grad():
+                pre_P2R_collected_tensors: Dict[str, torch.Tensor] = (
+                    self.pre_P2R_collect_parameters()
+                )
 
-                        view = (
-                            local_view.cosmos_slice(tensor_split_strategys)
-                            .contiguous()
-                            .cuda()
-                        )
-                        assert self.global_rank == p_rank
+                def grouped_send(grouped_send_ops):
+                    nccl_group_start(comm_id)
+                    for view, r_rank in grouped_send_ops:
                         nccl_send(
                             view,
                             self.world_size + r_rank,
                             comm_id,
                         )
-                        total_bytes_sent += view.numel() * view.element_size()
+                    nccl_group_end(comm_id)
+                    grouped_send_ops.clear()
+
+                grouped_send_ops = []
+                num_groups = 0
+
+                TRANSFER_GROUP_SIZE = 4
+                for insts_group in self.policy_to_rollout_insts:
+                    for insts_for_per_param in insts_group.param_instructions:
+                        dest_name = insts_for_per_param.param_name
+                        for inst in insts_for_per_param.instructions:
+                            p_rank = inst.policy_rank
+                            r_rank = inst.rollout_rank
+                            tensor_split_strategys = inst.slice_strategy
+                            if dest_name not in self.map_w_from_policy_to_rollout:
+                                raise RuntimeError(
+                                    f"dest_name {dest_name} not in map_w_from_policy_to_rollout"
+                                )
+                            local_view = self.map_w_from_policy_to_rollout[dest_name]
+                            if dest_name in pre_P2R_collected_tensors:
+                                local_view = pre_P2R_collected_tensors[dest_name]
+                            elif isinstance(local_view, Callable):
+                                local_view = local_view()
+                            else:
+                                pass
+
+                            view = (
+                                local_view.cosmos_slice(tensor_split_strategys)
+                                .contiguous()
+                                .cuda()
+                            )
+                            assert self.global_rank == p_rank
+                            logger.debug(
+                                f"Sending {dest_name} to rollout rank {r_rank}, {view.shape}"
+                            )
+                            grouped_send_ops.append((view, r_rank))
+                            total_bytes_sent += view.numel() * view.element_size()
+                    num_groups += 1
+                    if num_groups == TRANSFER_GROUP_SIZE:
+                        grouped_send(grouped_send_ops)
+                        num_groups = 0
+
+                grouped_send(grouped_send_ops)
+
         # make sure all the send operations of all ranks are finished
         time_eclapsed = time.time() - st
         logger.debug(
