@@ -1269,17 +1269,13 @@ class ParallelizedShardMapper:
                                 p_rank, r, p_tensor_split_strategys
                             ).__dict__
                         )
-            if insts_for_param_name:
-                insts_for_group.append(
-                    WeightSyncInstructionsPerParam(
-                        dest_name, insts_for_param_name
-                    ).__dict__
-                )
-            if insts_for_group:
-                policy_to_rollout_insts.append(
-                    WeightSyncInstructionsGroup(insts_for_group).__dict__
-                )
-            else:
+            insts_for_group.append(
+                WeightSyncInstructionsPerParam(dest_name, insts_for_param_name).__dict__
+            )
+            policy_to_rollout_insts.append(
+                WeightSyncInstructionsGroup(insts_for_group).__dict__
+            )
+            if not insts_for_param_name:
                 logger.warning(
                     f"No send instructions generated for parameter {dest_name} in policy rank {p_rank}."
                 )
@@ -1329,16 +1325,14 @@ class ParallelizedShardMapper:
                                     p_rank, r, p_tensor_split_strategys
                                 ).__dict__
                             )
-                if insts_for_param_name:
-                    insts_for_group.append(
-                        WeightSyncInstructionsPerParam(
-                            dest_name, insts_for_param_name
-                        ).__dict__
-                    )
-            if insts_for_group:
-                policy_to_rollout_insts.append(
-                    WeightSyncInstructionsGroup(insts_for_group).__dict__
+                insts_for_group.append(
+                    WeightSyncInstructionsPerParam(
+                        dest_name, insts_for_param_name
+                    ).__dict__
                 )
+            policy_to_rollout_insts.append(
+                WeightSyncInstructionsGroup(insts_for_group).__dict__
+            )
         if len(name_in_group) > 0:
             logger.warning(
                 f"No send instructions generated for parameters {name_in_group} in policy rank {p_rank}."
